#!/usr/local/bin/python
# encoding: utf-8
"""
*Download resources from a list of URLs.

There are options to rename all the downloaded resource, index the files, set differing download locations and pass basic authentication credentials.*

:Author:
    David Young
"""
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import str
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
import urllib
import requests
from multiprocessing.pool import ThreadPool
gtimeout = 10.
llog = ""
import random
import time


def multiobject_download(
    urlList,
    downloadDirectory,
    log,
    timeStamp=True,
    timeout=180,
    concurrentDownloads=10,
    resetFilename=False,
    credentials=False,
    longTime=False,
    indexFilenames=False
):
    """
    *get multiple url documents and place them in specified download directory/directories*

    **Key Arguments**

      - ``urlList`` -- list of document urls
      - ``downloadDirectory`` -- directory(ies) to download the documents to - can be one directory path or a list of paths the same length as urlList
      - ``log`` -- the logger
      - ``timestamp`` -- append a timestamp the name of the URL (ensure unique filenames)
      - ``longTime`` -- use a longer timestamp when appending to the filename (greater uniqueness)
      - ``timeout`` -- the timeout limit for downloads (secs)
      - ``concurrentDownloads`` -- the number of concurrent downloads allowed at any one time
      - ``resetFilename`` -- a string to reset all filenames to
      - ``credentials`` -- basic http credentials { 'username' : "...", "password", "..." }
      - ``indexFilenames`` -- prepend filenames with index (where url appears in urllist)

    **Return**

      - list of timestamped documents (same order as the input urlList)

    **Usage**

    ```python
    # download the pages linked from the main list page
    from fundamentals.download import multiobject_download
    localUrls = multiobject_download(
        urlList=["https://www.python.org/dev/peps/pep-0257/","https://en.wikipedia.org/wiki/Docstring"],
        downloadDirectory="/tmp",
        log="log",
        timeStamp=True,
        timeout=180,
        concurrentDownloads=2,
        resetFilename=False,
        credentials=False,  # { 'username' : "...", "password", "..." }
        longTime=True
    )

    print localUrls
    # OUT: ['/tmp/untitled_20160316t160650610780.html',
    # '/tmp/Docstring_20160316t160650611136.html']
    ```

    .. image:: https://i.imgur.com/QYoMm24.png width=600px

    """
    import sys
    import os
    import re
    from fundamentals.download import append_now_datestamp_to_filename, extract_filename_from_url

    # TIMEOUT IN SECONDS
    global gtimeout
    global llog
    llog = log
    gtimeout = float(timeout)

    # BUILD THE 2D ARRAY FOR MULTI_THREADED DOWNLOADS
    thisArray = []
    bodies = []
    localUrls = []
    theseUrls = []
    requestList = []

    totalCount = len(urlList)

    # IF ONLY ONE DOWNLOAD DIRECORY
    if not isinstance(downloadDirectory, list):
        for i, url in enumerate(urlList):
            # EXTRACT THE FILENAME FROM THE URL
            if resetFilename and len(resetFilename):
                filename = resetFilename[i]
            else:
                filename = extract_filename_from_url(log, url)
                if indexFilenames:
                    filename = """%(i)03d_%(filename)s""" % locals()

            if not filename:
                from datetime import datetime, date, time
                now = datetime.now()
                filename = now.strftime("%Y%m%dt%H%M%S%f")

            if(timeStamp):
                # APPEND TIMESTAMP TO THE FILENAME
                filename = append_now_datestamp_to_filename(
                    log, filename, longTime=longTime)
            # GENERATE THE LOCAL FILE URL
            localFilepath = downloadDirectory + "/" + filename

            # ADD BASIC AUTH TO THE URLS
            if credentials != False:
                url_pass = f'{credentials["username"]}:{credentials["password"]}@'
                if "://" in url:
                    url = url.replace("://", "://" + url_pass)
                else:
                    url = url_pass + url
            thisArray.extend([[url, localFilepath]])

    else:
        for url, d in zip(urlList, downloadDirectory):
            # EXTRACT THE FILENAME FROM THE URL
            if resetFilename:
                filename = resetFilename
            else:
                filename = extract_filename_from_url(log, url)

            if not filename:
                continue

            if(timeStamp):
                # APPEND TIMESTAMP TO THE FILENAME
                filename = append_now_datestamp_to_filename(
                    log, filename)
            # GENERATE THE LOCAL FILE URL
            localFilepath = d + "/" + filename
            thisArray.extend([[url, localFilepath]])

            # ADD BASIC AUTH TO THE URLS
            if credentials != False:
                url_pass = f'{credentials["username"]}:{credentials["password"]}@'
                if "://" in url:
                    url = url.replace("://", "://" + url_pass)
                else:
                    url = url_pass + url

    # CONCURRENTLY DOWNLOAD URLS
    results = ThreadPool(concurrentDownloads).imap_unordered(
        fetch_url, thisArray)
    urlNum = 0
    returnPaths = []
    for path in results:
        returnPaths.append(path)
        urlNum += 1
        if urlNum > 1:
            # CURSOR UP ONE LINE AND CLEAR LINE
            sys.stdout.write("\x1b[1A\x1b[2K")
        percent = (float(urlNum) / float(totalCount)) * 100.
        print("  %(urlNum)s / %(totalCount)s (%(percent)1.1f%%) URLs downloaded" % locals())

    localPaths = []
    localPaths[:] = [o[1] for o in thisArray if o[1] in returnPaths]

    return localPaths


def fetch_url(entry):
    downloaded = False
    tries = 5
    count = 0
    uri, path = entry
    timeout = gtimeout

    randSleep = random.randint(1, 101) / 20.

    time.sleep(randSleep)

    while not downloaded and count < tries:
        try:
            r = requests.get(uri, stream=True, timeout=timeout)
        except:
            count += 1
            timeout *= 2
            llog.warning(f"timeout on attempt number {count}/{tries}. Increasing to {timeout}s")
            continue

        if r.status_code == 200:
            with open(path, 'wb') as f:
                for chunk in r:
                    f.write(chunk)
            return path
        else:
            count += 1
            llog.warning(f"Getting status code {r.status_code} on download attempt {count}/{tries}.")
            downloaded = False

    return None



#!/usr/local/bin/python
# encoding: utf-8
"""
*Given a mysql query, read the data from the database and return the results as a list of dictionaries (database rows)*

:Author:
    David Young
"""
from builtins import str
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools


def readquery(
        sqlQuery,
        dbConn,
        log,
        quiet=False):
    """Given a mysql query, read the data from the database and return the results as a list of dictionaries (database rows)

    **Key Arguments**

    - ``log`` -- the logger.
    - ``sqlQuery`` -- the MySQL command to execute
    - ``dbConn`` -- the db connection
    - ``quiet`` -- ignore mysql warnings and errors and move on. Be careful when setting this to true - damaging errors can easily be missed. Default *False*.


    **Return**

    - ``rows`` -- the rows returned by the sql query


    **Usage**

    ```python
    from fundamentals.mysql import readquery
    rows = readquery(
        log=log,
        sqlQuery=sqlQuery,
        dbConn=dbConn,
        quiet=False
    )
    ```

    """
    log.debug('starting the ``readquery`` function')
    import pymysql
    import warnings
    warnings.filterwarnings('error', category=pymysql.Warning)

    rows = []

    log.debug("\nSQLQUERY: %(sqlQuery)s}\n" % locals())

    try:
        cursor = dbConn.cursor(pymysql.cursors.DictCursor)
    except Exception as e:
        log.error('could not create the database cursor: %s' % (e, ))
        raise IOError('could not create the database cursor: %s' % (e, ))
    # EXECUTE THE SQL COMMAND
    tryAgain = True
    tries = 1
    while tryAgain:
        tryAgain = False
        try:
            cursor.execute(sqlQuery)
            rows = cursor.fetchall()
        except pymysql.err.InternalError as e:
            if tries < 61:
                tryAgain = True
                log.warning(f"MySQL error: {e}. Attempt {tries}/60.")
                tries += 1
            else:
                log.warning(f"MySQL error: {e}. Attempt {tries}/60 failed. ")
                raise
        except Exception as e:
            sqlQuery = sqlQuery[:1000]
            if quiet == False:
                log.warning(
                    'MySQL raised an error - read command not executed.\n' + str(e) + '\nHere is the sqlQuery\n\t%(sqlQuery)s' % locals())
                raise e
            else:
                log.warning(
                    'MySQL raised an error - read command not executed.\n' + str(e) + '\nHere is the sqlQuery\n\t%(sqlQuery)s' % locals())
                pass

    # CLOSE THE CURSOR
    try:
        cursor.close()
    except Exception as e:
        log.warning('could not close the db cursor ' + str(e) + '\n')

    log.debug('completed the ``readquery`` function')
    return rows



#!/usr/bin/env python
# encoding: utf-8
"""
*Toolset to setup the main function for a cl-util*

:Author:
    David Young
"""
from __future__ import print_function
from __future__ import absolute_import
from builtins import object
import sys
import os
import yaml
try:
    yaml.warnings({'YAMLLoadWarning': False})
except:
    pass
from collections import OrderedDict
import shutil
from subprocess import Popen, PIPE, STDOUT
from . import logs as dl
import time
from docopt import docopt
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
from os.path import expanduser
import inspect
import psutil


###################################################################
# CLASSES                                                         #
###################################################################


class tools(object):
    """
    *common setup methods & attributes of the main function in cl-util*

    **Key Arguments**

    - ``dbConn`` -- mysql database connection
    - ``arguments`` -- the arguments read in from the command-line
    - ``docString`` -- pass the docstring from the host module so that docopt can work on the usage text to generate the required arguments
    - ``logLevel`` -- the level of the logger required. Default *DEBUG*. [DEBUG|INFO|WARNING|ERROR|CRITICAL]
    - ``options_first`` -- options come before commands in CL usage. Default *False*.
    - ``projectName`` -- the name of the project, used to create a default settings file in ``~/.config/projectName/projectName.yaml``. Default *False*.
    - ``distributionName`` -- the distribution name if different from the projectName (i.e. if the package is called by another name on PyPi). Default *False*
    - ``tunnel`` -- will setup a ssh tunnel (if the settings are found in the settings file). Default *False*.
    - ``defaultSettingsFile`` -- if no settings file is passed via the doc-string, look for a settings file first in the PWD or use the default settings file in ``~/.config/projectName/projectName.yaml`` (don't have to clutter command-line with settings)

    **Usage**

    Add this to the ``__main__`` function of your command-line module

    ```python
    # setup the command-line util settings
    from fundamentals import tools
    su = tools(
        arguments=arguments,
        docString=__doc__,
        logLevel="DEBUG",
        options_first=False,
        projectName="myprojectName"
    )
    arguments, settings, log, dbConn = su.setup()
    ```

    Here is a template settings file content you could use:

    ```yaml
    version: 1
    database settings:
        db: unit_tests
        host: localhost
        user: utuser
        password: utpass
        tunnel: true

    # SSH TUNNEL - if a tunnel is required to connect to the database(s) then add setup here
    # Note only one tunnel is setup - may need to change this to 2 tunnels in the future if 
    # code, static catalogue database and transient database are all on seperate machines.
    ssh tunnel:
        remote user: username
        remote ip: mydomain.co.uk
        remote datbase host: mydatabaseName
        port: 9002

    logging settings:
        formatters:
            file_style:
                format: '* %(asctime)s - %(name)s - %(levelname)s (%(pathname)s > %(funcName)s > %(lineno)d) - %(message)s  '
                datefmt: '%Y/%m/%d %H:%M:%S'
            console_style:
                format: '* %(asctime)s - %(levelname)s: %(pathname)s:%(funcName)s:%(lineno)d > %(message)s'
                datefmt: '%H:%M:%S'
            html_style:
                format: '<div id="row" class="%(levelname)s"><span class="date">%(asctime)s</span>   <span class="label">file:</span><span class="filename">%(filename)s</span>   <span class="label">method:</span><span class="funcName">%(funcName)s</span>   <span class="label">line#:</span><span class="lineno">%(lineno)d</span> <span class="pathname">%(pathname)s</span>  <div class="right"><span class="message">%(message)s</span><span class="levelname">%(levelname)s</span></div></div>'
                datefmt: '%Y-%m-%d <span class= "time">%H:%M <span class= "seconds">%Ss</span></span>'
        handlers:
            console:
                class: logging.StreamHandler
                level: DEBUG
                formatter: console_style
                stream: ext://sys.stdout
            file:
                class: logging.handlers.GroupWriteRotatingFileHandler
                level: WARNING
                formatter: file_style


                    filename: /Users/Dave/.config/myprojectName/myprojectName.log
                    mode: w+
                    maxBytes: 102400
                    backupCount: 1
            root:
                level: WARNING
                handlers: [file,console]
        ```
    """
    # Initialisation

    def __init__(
            self,
            arguments,
            docString,
            logLevel="WARNING",
            options_first=False,
            projectName=False,
            distributionName=False,
            orderedSettings=False,
            defaultSettingsFile=False,
            quitIfRunning=True
    ):
        self.arguments = arguments
        self.docString = docString
        self.logLevel = logLevel
        self.configSettingsPath = os.getenv("HOME") + f"/.config/{projectName}/{projectName}.yaml"
        self.projectName = projectName

        if not distributionName:
            distributionName = projectName

        version = '0.0.1'
        try:
            import pkg_resources
            version = pkg_resources.get_distribution(distributionName).version
        except:
            pass

        ## ACTIONS BASED ON WHICH ARGUMENTS ARE RECIEVED ##
        # PRINT COMMAND-LINE USAGE IF NO ARGUMENTS PASSED
        if self.arguments == None:
            self.arguments = docopt(docString, version="v" + version,
                                    options_first=options_first)

        # BUILD A STRING FOR THE PROCESS TO MATCH RUNNING PROCESSES AGAINST
        lockname = "".join(sys.argv)

        # TEST IF THE PROCESS IS ALREADY RUNNING WITH THE SAME ARGUMENTS (e.g.
        # FROM CRON) - QUIT IF MATCH FOUND
        if quitIfRunning:
            for q in psutil.process_iter():
                try:
                    this = q.cmdline()
                except:
                    continue

                test = "".join(this[1:])
                if q.pid != os.getpid() and lockname == test and "--reload" not in test:
                    thisId = q.pid
                    print("This command is already running (see PID %(thisId)s)" % locals())
                    sys.exit(0)

        try:
            if "tests.test" in self.arguments["<pathToSettingsFile>"]:
                del self.arguments["<pathToSettingsFile>"]
        except:
            pass

        # GET ADVANCED SETTINGS IF AVAILABLE
        advs = os.getcwd() + "/rubbish.yaml"
        level = 0
        exists = False
        count = 1
        while not exists and len(advs) and count < 10:
            count += 1
            level -= 1
            exists = os.path.exists(advs)
            if not exists:
                advs = "/".join(inspect.stack()
                                [1][1].split("/")[:level]) + "/advanced_settings.yaml"
        if not exists:
            advs = {}
        else:
            with open(advs, 'r') as stream:
                advs = yaml.safe_load(stream)

        if defaultSettingsFile and "settingsFile" not in self.arguments and "--settings" not in self.arguments:
            cwdSettings = os.getcwd() + f"/{projectName}.yaml"
            if os.path.exists(cwdSettings):
                self.arguments["settingsFile"] = settingsFile = cwdSettings
            elif os.path.exists(os.getenv(
                    "HOME") + "/.config/%(projectName)s/%(projectName)s.yaml" % locals()):
                self.arguments["settingsFile"] = settingsFile = os.getenv(
                    "HOME") + "/.config/%(projectName)s/%(projectName)s.yaml" % locals()

        # UNPACK SETTINGS
        stream = False
        if "<settingsFile>" in self.arguments and self.arguments["<settingsFile>"]:
            stream = open(self.arguments["<settingsFile>"], 'r')
        elif "<pathToSettingsFile>" in self.arguments and self.arguments["<pathToSettingsFile>"]:
            stream = open(self.arguments["<pathToSettingsFile>"], 'r')
        elif "--settingsFile" in self.arguments and self.arguments["--settingsFile"]:
            stream = open(self.arguments["--settingsFile"], 'r')
        elif "--settings" in self.arguments and self.arguments["--settings"]:
            stream = open(self.arguments["--settings"], 'r')
        elif "pathToSettingsFile" in self.arguments and self.arguments["pathToSettingsFile"]:
            stream = open(self.arguments["pathToSettingsFile"], 'r')
        elif "settingsFile" in self.arguments and self.arguments["settingsFile"]:
            stream = open(self.arguments["settingsFile"], 'r')
        elif ("settingsFile" in self.arguments and self.arguments["settingsFile"] == None) or ("<pathToSettingsFile>" in self.arguments and self.arguments["<pathToSettingsFile>"] == None) or ("--settings" in self.arguments and self.arguments["--settings"] == None) or ("pathToSettingsFile" in self.arguments and self.arguments["pathToSettingsFile"] == None):

            init = False
            workspaceDirectory = False
            if "init" in self.arguments and self.arguments["init"]:
                init = True
                if "<workspaceDirectory>" in self.arguments and self.arguments["<workspaceDirectory>"]:
                    theseSettings = self.arguments["<workspaceDirectory>"] + f"/{projectName}.yaml"
                else:
                    theseSettings = self.configSettingsPath
                exists = self._create_or_verify_settings(pathToSettings=theseSettings, create=True)

            else:
                if projectName != False:
                    # FIRST CHECK FOR SETTINGS IN CWD
                    cwdSettings = os.getcwd() + f"/{projectName}.yaml"
                    exists = self._create_or_verify_settings(pathToSettings=cwdSettings, create=False)

                    # THEN CHECK FOR SETTINGS CONFIG DIRECTORY
                    if not exists:
                        exists = self._create_or_verify_settings(pathToSettings=self.configSettingsPath, create=False)

                    # CREATE SETTING IN DEFAULT SETTINGS LOCATION IF NONE EXIST
                    if not exists:
                        exists = self._create_or_verify_settings(pathToSettings=self.configSettingsPath, create=True)
            stream = open(self.arguments["settingsFile"], 'r')
        else:
            pass

        # FOR SETTINGS FILE PATHS PASSED DIRECTORY VIA THE CL
        if stream is not False:
            astream = stream.read()
            home = expanduser("~")
            astream = astream.replace("~/", home + "/")

            if orderedSettings:
                settings = ordered_load(astream, yaml.SafeLoader)
            else:
                settings = yaml.safe_load(astream)

        # MERGE ADVANCED SETTINGS AND USER SETTINGS (USER SETTINGS OVERRIDE)
        if 'settings' not in locals():
            settings = advs
        else:
            settings = {**advs, **settings}

        # SETUP LOGGER -- DEFAULT TO CONSOLE LOGGER IF NONE PROVIDED IN
        # SETTINGS
        if 'settings' in locals() and "logging settings" in settings:
            if "settingsFile" in self.arguments:
                log = dl.setup_dryx_logging(
                    yaml_file=self.arguments["settingsFile"]
                )
            elif "<settingsFile>" in self.arguments:
                log = dl.setup_dryx_logging(
                    yaml_file=self.arguments["<settingsFile>"]
                )
            elif "<pathToSettingsFile>" in self.arguments:
                log = dl.setup_dryx_logging(
                    yaml_file=self.arguments["<pathToSettingsFile>"]
                )
            elif "--settingsFile" in self.arguments:
                log = dl.setup_dryx_logging(
                    yaml_file=self.arguments["--settingsFile"]
                )
            elif "pathToSettingsFile" in self.arguments:
                log = dl.setup_dryx_logging(
                    yaml_file=self.arguments["pathToSettingsFile"]
                )

            elif "--settings" in self.arguments:
                log = dl.setup_dryx_logging(
                    yaml_file=self.arguments["--settings"]
                )

        elif "--logger" not in self.arguments or self.arguments["--logger"] is None:
            log = dl.console_logger(
                level=self.logLevel
            )

        self.log = log

        # unpack remaining cl arguments using `exec` to setup the variable names
        # automatically
        for arg, val in list(self.arguments.items()):
            if arg[0] == "-":
                varname = arg.replace("-", "") + "Flag"
            else:
                varname = arg.replace("<", "").replace(">", "")
            if varname == "import":
                varname = "iimport"
            if isinstance(val, str):
                val = val.replace("'", "\\'")
                exec(varname + " = '%s'" % (val,))
            else:
                exec(varname + " = %s" % (val,))
            if arg == "--dbConn":
                dbConn = val

        # SETUP A DATABASE CONNECTION BASED ON WHAT ARGUMENTS HAVE BEEN PASSED
        dbConn = False
        tunnel = False

        if ("--host" in self.arguments and "--dbName" in self.arguments and self.arguments["--host"]):
            # SETUP DB CONNECTION
            dbConn = True
            host = self.arguments["--host"]
            user = self.arguments["--user"]
            passwd = self.arguments["--passwd"]
            dbName = self.arguments["--dbName"]
            port = False

        elif 'settings' in locals() and "database settings" in settings and "host" in settings["database settings"]:
            host = settings["database settings"]["host"]
            user = settings["database settings"]["user"]
            passwd = settings["database settings"]["password"]
            dbName = settings["database settings"]["db"]
            if "tunnel" in settings["database settings"] and settings["database settings"]["tunnel"]:
                tunnel = True
            dbConn = True
            port = False
            if "port" in settings["database settings"] and settings["database settings"]["port"]:
                port = int(settings["database settings"]["port"])
        else:
            pass

        if not 'settings' in locals():
            settings = False
        self.settings = settings

        if tunnel:
            self._setup_tunnel()
            self.dbConn = self.remoteDBConn
            return None

        if dbConn:
            import pymysql as ms
            dbConn = ms.connect(
                host=host,
                user=user,
                passwd=passwd,
                db=dbName,
                port=port,
                use_unicode=True,
                charset='utf8',
                local_infile=1,
                client_flag=ms.constants.CLIENT.MULTI_STATEMENTS,
                connect_timeout=36000,
                max_allowed_packet=51200000
            )
            dbConn.autocommit(True)

        self.dbConn = dbConn

        return None

    def setup(
            self):
        """
        **Summary:**
            *setup the attributes and return*
        """

        return self.arguments, self.settings, self.log, self.dbConn

    def _setup_tunnel(
            self):
        """
        *setup ssh tunnel if required*
        """
        from subprocess import Popen, PIPE, STDOUT
        import pymysql as ms

        # SETUP TUNNEL IF REQUIRED
        if "ssh tunnel" in self.settings:
            # TEST TUNNEL DOES NOT ALREADY EXIST
            sshPort = self.settings["ssh tunnel"]["port"]
            connected = self._checkServer(
                self.settings["database settings"]["host"], sshPort)
            if connected:
                pass
            else:
                # GRAB TUNNEL SETTINGS FROM SETTINGS FILE
                ru = self.settings["ssh tunnel"]["remote user"]
                rip = self.settings["ssh tunnel"]["remote ip"]
                rh = self.settings["ssh tunnel"]["remote datbase host"]

                cmd = "ssh -fnN %(ru)s@%(rip)s -L %(sshPort)s:%(rh)s:3306" % locals()
                p = Popen(cmd, shell=True, close_fds=True)
                output = p.communicate()[0]

                # TEST CONNECTION - QUIT AFTER SO MANY TRIES
                connected = False
                count = 0
                while not connected:
                    connected = self._checkServer(
                        self.settings["database settings"]["host"], sshPort)
                    time.sleep(1)
                    count += 1
                    if count == 5:
                        self.log.error(
                            'cound not setup tunnel to remote datbase' % locals())
                        sys.exit(0)

        if "tunnel" in self.settings["database settings"] and self.settings["database settings"]["tunnel"]:
            # TEST TUNNEL DOES NOT ALREADY EXIST
            sshPort = self.settings["database settings"]["tunnel"]["port"]
            connected = self._checkServer(
                self.settings["database settings"]["host"], sshPort)
            if connected:
                pass
            else:
                # GRAB TUNNEL SETTINGS FROM SETTINGS FILE
                ru = self.settings["database settings"][
                    "tunnel"]["remote user"]
                rip = self.settings["database settings"]["tunnel"]["remote ip"]
                rh = self.settings["database settings"][
                    "tunnel"]["remote datbase host"]

                cmd = "ssh -fnN %(ru)s@%(rip)s -L %(sshPort)s:%(rh)s:3306" % locals()
                p = Popen(cmd, shell=True, close_fds=True)
                output = p.communicate()[0]

                # TEST CONNECTION - QUIT AFTER SO MANY TRIES
                connected = False
                count = 0
                while not connected:
                    connected = self._checkServer(
                        self.settings["database settings"]["host"], sshPort)
                    time.sleep(1)
                    count += 1
                    if count == 5:
                        self.log.error(
                            'cound not setup tunnel to remote datbase' % locals())
                        sys.exit(0)

        # SETUP A DATABASE CONNECTION FOR THE remote database
        host = self.settings["database settings"]["host"]
        user = self.settings["database settings"]["user"]
        passwd = self.settings["database settings"]["password"]
        dbName = self.settings["database settings"]["db"]
        thisConn = ms.connect(
            host=host,
            user=user,
            passwd=passwd,
            db=dbName,
            port=sshPort,
            use_unicode=True,
            charset='utf8',
            local_infile=1,
            client_flag=ms.constants.CLIENT.MULTI_STATEMENTS,
            connect_timeout=36000,
            max_allowed_packet=51200000
        )
        thisConn.autocommit(True)
        self.remoteDBConn = thisConn

        return None

    def _checkServer(self, address, port):
        """
        *Check that the TCP Port we've decided to use for tunnelling is available*
        """
        # CREATE A TCP SOCKET
        import socket
        s = socket.socket()

        try:
            s.connect((address, port))
            return True
        except socket.error as e:
            self.log.warning(
                """Connection to `%(address)s` on port `%(port)s` failed - try again: %(e)s""" % locals())
            return False

        return None

    def _create_or_verify_settings(
            self,
            pathToSettings,
            create=False):
        """*create and/or verify a settings file at a given path*

        **Key Arguments:**
            - ``pathToSettings`` -- path to create/verify
            - ``create`` -- create the file if it does not exist. Default *False*.

        **Return:**
            - ``exists`` -- does the setting file now exist?
        """

        import codecs

        exists = False

        absPath = os.path.abspath(pathToSettings)
        defaultLog = f"{os.getenv('HOME')}/.config/{self.projectName}/{self.projectName}.log".replace("//", "/")
        alternativeLogPath = absPath.replace(".yaml", ".log")

        if os.path.exists(pathToSettings):
            self.arguments["settingsFile"] = pathToSettings
            exists = True

        if exists and create:
            print(f"A settings file already exists at '{pathToSettings}' and has not been modified.")

        if not exists and create:
            # FIND THE DEAFULT SETTINGS FILE
            ds = os.getcwd() + "/rubbish.yaml"
            level = 0
            exists = False
            count = 1
            while not exists and len(ds) and count < 10:
                count += 1
                level -= 1
                exists = os.path.exists(ds)
                if not exists:

                    ds = "/".join(inspect.stack()[2][1].split("/")[:level]) + "/default_settings.yaml"

            try:
                # COPY THE SETTINGS FILE TO THE REQUESTED LOCATION
                shutil.copyfile(ds, pathToSettings)
                try:
                    readFile = codecs.open(
                        pathToSettings, encoding='utf-8', mode='r')
                    thisData = readFile.read()
                    readFile.close()
                except IOError as e:
                    message = 'could not open the file %s' % (
                        pathToReadFile,)
                    raise IOError(message)

                # JUST INCASE!
                thisData = thisData.replace(
                    "/Users/Dave", os.getenv("HOME"))

                if defaultLog != alternativeLogPath:
                    thisData = thisData.replace(defaultLog, alternativeLogPath)

                # REWRITE CLEANED FILE
                try:
                    writeFile = codecs.open(
                        pathToSettings, encoding='utf-8', mode='w')
                except IOError as e:
                    message = 'could not open the file %s' % (
                        pathToSettings,)
                    raise IOError(message)
                writeFile.write(thisData)
                writeFile.close()
                print(
                    f"Default settings have been added to '{pathToSettings}'. Tailor these settings before proceeding to run {self.projectName}.")
            except:
                print(
                    f"Please add any require settings to file '{pathToSettings}' before proceeding to run {self.projectName}.")
            self.arguments["settingsFile"] = pathToSettings

        return exists

    # use the tab-trigger below for new method
    # xt-class-method


def ordered_load(stream, Loader=yaml.loader, object_pairs_hook=OrderedDict):
    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))
    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)
    return yaml.safe_load(stream, OrderedLoader)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# encoding: utf-8
"""
*helper class to daemonise you code*

:Author:
    David Young

:Date Created:
    May 23, 2023
"""

from fundamentals import tools
from builtins import object
import sys
import os
os.environ['TERM'] = 'vt100'


class daemonise():
    """
    A class to daemonise a python code

    **Key Arguments:**
        - ``log`` -- logger
        - ``name`` -- a name for the daemon (e.g. python package name)

    **Usage:**

    Add something like this to your command-line usage:

    ```text
    Usage:
    myCommand (start|stop|restart|status)

    Options:
        start                 start the myCommand daemon
        stop                  stop the myCommand daemon
        restart               restart the myCommand daemon
        status                print the staus of the myCommand daemon 
    ```

    and then when executing the commands:

    ```python
    from fundamental import daemonise
    class myDaemon(daemonise):
        def action(
                self,
                **kwargs):
            self.log.info('starting the ``action`` method')

            anotherParameter = kwargs["anotherParameter"]

            import time
            while True:
                print(f"OVERRIDE ACTION - {anotherParameter}")
                time.sleep(3)

            self.log.info('completed the ``action`` method')
            return None

    d = myDaemon(log=log, name="gocart", anotherParameter=42.)

    if a['start']:
        d.start()
    elif a['stop']:
        d.stop()
    elif a['restart']:
        d.restart()
    elif a['status']:
        d.status()
    ```

    Replace `**akws` with any keywords you need.

    """

    def __init__(
            self,
            log,
            name,
            **akws
    ):
        self.log = log
        log.debug("instansiating a new 'daemonise' object")

        # MAKE ROOT DIR
        from os.path import expanduser
        home = expanduser("~")
        self.rootDir = home + "/.config/gocart/"
        if not os.path.exists(self.rootDir):
            os.makedirs(self.rootDir)

        self.name = name
        self.errLog = home + f"/.config/{name}/daemon_err.log"
        self.outLog = home + f"/.config/{name}/daemon.log"
        self.pidFile = home + f"/.config/{name}/daemon.pid"
        self.akws = akws

        return

    def start(self):
        """start the daemonise running
        """
        self.log.info('starting the ``get`` method')

        import time
        import signal
        from daemon import pidfile
        import daemon

        if os.path.exists(self.pidFile):
            with open(self.pidFile, mode='r') as f:
                pid = f.read().strip()
                print(f"{self.name} is already running (PID = {pid}).")
        else:
            print(f"{self.name} has been started.")
            print(f"The logs can be found here: {self.rootDir}")
            with daemon.DaemonContext(
                working_directory=self.rootDir,
                umask=0o002,
                pidfile=pidfile.TimeoutPIDLockFile(self.pidFile),
                stdout=open(self.outLog, 'a'),
                stderr=open(self.errLog, 'a'),
                signal_map={signal.SIGTERM: self.cleanup}
            ) as context:
                self.action(**self.akws)

        self.log.info('completed the ``get`` method')
        return None

    def cleanup(
            self,
            signum,
            frame):
        """*the code to run when daemon is killed*
        """
        self.log.debug('starting the ``cleanup`` method')

        from datetime import datetime, date, time
        now = datetime.now()
        now = now.strftime("%Y%m%dt%H%M%S")

        print(f"{self.name} stopped at {now}")
        sys.exit(0)

        self.log.debug('completed the ``cleanup`` method')
        return None

    def action(
            self,
            **akws):
        """*the code to execute in daemon mode, this method should be overriden to execute novel code*
        """
        self.log.debug('starting the ``action`` method')

        import time
        while True:
            print("ACTION")
            time.sleep(3)

        self.log.debug('completed the ``action`` method')
        return None

    def stop(
            self):
        """*stop the daemon and cleanup*
        """
        self.log.debug('starting the ``stop`` method')

        import signal

        if os.path.exists(self.pidFile):
            with open(self.pidFile, mode='r') as f:
                pid = f.read().strip()
            os.kill(int(pid), signal.SIGTERM)
            print(f"{self.name} has been stopped.")
        else:
            print(f"{self.name} is not running.")

        self.log.debug('completed the ``stop`` method')
        return None

    def status(
            self):
        """*print the status of the daemon*
        """
        self.log.debug('starting the ``status`` method')

        if os.path.exists(self.pidFile):
            with open(self.pidFile, mode='r') as f:
                pid = f.read().strip()
            print(f"{self.name} is running (PID = {pid}).")
        else:
            print(f"{self.name} is not running.")

        self.log.debug('completed the ``status`` method')
        return None

    def restart(
            self):
        """*stop and start the daemon*
        """
        self.log.debug('starting the ``status`` method')

        import time
        self.stop()

        iteration = 0
        while os.path.exists(self.pidFile) and iteration < 15:
            iteration += 1
            time.sleep(2)

        self.start()

        self.log.debug('completed the ``status`` method')
        return None



#!/usr/local/bin/python
# encoding: utf-8
"""
*A function to quickly add multiprocessing to any program*

:Author:
    David Young
"""
from __future__ import division
import inspect
from functools import partial
from fundamentals import tools
from past.utils import old_div
import sys
import os
os.environ['TERM'] = 'vt100'


def fmultiprocess(
        log,
        function,
        inputArray,
        poolSize=False,
        timeout=3600,
        turnOffMP=False,
        **kwargs):
    """multiprocess pool

    **Key Arguments**

    - ``log`` -- logger
    - ``function`` -- the function to multiprocess
    - ``inputArray`` -- the array to be iterated over
    - ``poolSize`` -- limit the number of CPU that are used in multiprocess job
    - ``timeout`` -- time in sec after which to raise a timeout error if the processes have not completed
    - ``turnOffMP`` -- turn off multiprocessing. Useful for profiling and debugging. Default **False**


    **Return**

    - ``resultArray`` -- the array of results


    **Usage**

    ```python
    from fundamentals import multiprocess
    # DEFINE AN INPUT ARRAY
    inputArray = range(10000)
    results = multiprocess(log=log, function=functionName, poolSize=10, timeout=300,
                          inputArray=inputArray, otherFunctionKeyword="cheese")
    ```

    """
    log.debug('starting the ``multiprocess`` function')

    logFound = False
    # PYTHON 3 VS 2 ..
    try:
        if "log" in inspect.getfullargspec(function)[0]:
            logFound = True
    except:
        if "log" in inspect.getargspec(function)[0]:
            logFound = True

    if turnOffMP == False:
        import psutil
        # import multiprocess as mp
        # mp.set_start_method('forkserver')
        from multiprocess import cpu_count, Pool

        # DEFINTE POOL SIZE - NUMBER OF CPU CORES TO USE (BEST = ALL - 1)
        if not poolSize:
            poolSize = psutil.cpu_count()

        if poolSize:
            p = Pool(processes=poolSize)
        else:
            p = Pool()

        cpuCount = psutil.cpu_count()
        chunksize = int(old_div((len(inputArray) + 1), (cpuCount * 3)))

        if chunksize == 0:
            chunksize = 1

        # chunksize = 1
        # MAP-REDUCE THE WORK OVER MULTIPLE CPU CORES
        if logFound:
            mapfunc = partial(function, log=log, **kwargs)
            resultArray = p.map_async(mapfunc, inputArray, chunksize=chunksize)
        else:
            mapfunc = partial(function, **kwargs)
            resultArray = p.map_async(mapfunc, inputArray, chunksize=chunksize)

        resultArray = resultArray.get(timeout=timeout)

        p.close()
        p.join()
        # p.terminate()

    else:
        resultArray = []

        if logFound:
            for i in inputArray:
                r = function(log, i, **kwargs)
                resultArray.append(r)
        else:
            for i in inputArray:
                r = function(i, **kwargs)
                resultArray.append(r)

    log.debug('completed the ``multiprocess`` function')
    return resultArray



#!/usr/local/bin/python
# encoding: utf-8
"""
*Execute a MySQL write query on a database table*

:Author:
    David Young
"""
from builtins import str
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
import time


def writequery(
    log,
    sqlQuery,
    dbConn,
    Force=False,
    manyValueList=False
):
    """*Execute a MySQL write command given a sql query*

    **Key Arguments**

    - ``sqlQuery`` -- the MySQL command to execute
    - ``dbConn`` -- the db connection
    - ``Force`` -- do not exit code if error occurs, move onto the next command
    - ``manyValueList`` -- a list of value tuples if executing more than one insert


    **Return**

    - ``message`` -- error/warning message


    **Usage**

    Here's an example of how to create a table using the database connection passed to the function:

    ```python
    from fundamentals.mysql import writequery
    sqlQuery = "CREATE TABLE `testing_table` (`id` INT NOT NULL, PRIMARY KEY (`id`))"
    message = writequery(
        log=log,
        sqlQuery=sqlQuery,
        dbConn=dbConn,
        Force=False,
        manyValueList=False
    )
    ```

    Here's a many value insert example:

    ```python
    from fundamentals.mysql import writequery
    sqlQuery = "INSERT INTO testing_table (id) values (%s)"
    message = writequery(
        log=log,
        sqlQuery=sqlQuery,
        dbConn=dbConn,
        Force=False,
        manyValueList=[(1,), (2,), (3,), (4,), (5,), (6,), (7,),
                       (8,), (9,), (10,), (11,), (12,), ]
    )
    ```

    """
    log.debug('starting the ``writequery`` function')
    import pymysql
    import warnings
    warnings.filterwarnings('error', category=pymysql.Warning)
    message = ""
    try:
        cursor = dbConn.cursor(pymysql.cursors.DictCursor)
    except Exception as e:
        log.error('could not create the database cursor.')
    # EXECUTE THE SQL COMMAND

    log.debug("\nSQLQUERY: %(sqlQuery)s}\n" % locals())
    if sqlQuery[-1] == ",":
        sqlQuery = sqlQuery[:-1]

    tryAgain = True
    tries = 1
    while tryAgain:
        tryAgain = False
        try:
            if manyValueList == False:
                cursor.execute(sqlQuery)
            else:
                # cursor.executemany(sqlQuery, manyValueList)
                # INSET LARGE LISTS IN BATCHES TO STOP MYSQL SERVER BARFING
                batch = 100000
                offset = 0
                stop = 0

                while stop == 0:
                    thisList = manyValueList[offset:offset + batch]
                    offset += batch
                    a = len(thisList)
                    cursor.executemany(sqlQuery, thisList)
                    dbConn.commit()
                    if len(thisList) < batch:
                        stop = 1
        except (pymysql.err.InternalError, pymysql.err.OperationalError) as e:
            if tries < 61:
                tryAgain = True
                log.warning(f"MySQL error: {e}. Attempt {tries}/60.")
                tries += 1
            else:
                log.warning(f"MySQL error: {e}. Attempt {tries}/60 failed. ")
                raise

        except pymysql.err.ProgrammingError as e:
            message = 'MySQL write command not executed for this query: << %s >>\nThe error was: %s \n' % (sqlQuery,
                                                                                                           str(e))
            if Force == False:
                log.error(message)
                raise
            else:
                log.warning(message)
        except pymysql.Error as e:
            try:
                e = e.args
            except:
                pass

            if e[0] == 1050 and 'already exists' in e[1]:
                log.info(str(e) + '\n')
            elif e[0] == 1062:
                               # Duplicate Key error
                log.debug('Duplicate Key error: %s\n' % (str(e), ))
                message = "duplicate key error"
            elif e[0] == 1061:
                               # Duplicate Key error
                log.debug('index already exists: %s\n' % (str(e), ))
                message = "index already exists"
            elif "Duplicate entry" in str(e):
                log.debug('Duplicate Key error: %s\n' % (str(e), ))
                message = "duplicate key error"
            elif "Deadlock" in str(e):
                i = 0
                while i < 10:
                    time.sleep(1)
                    i += 1
                    try:
                        if manyValueList == False:
                            cursor.execute(sqlQuery)
                        else:
                            # cursor.executemany(sqlQuery, manyValueList)
                            # INSET LARGE LISTS IN BATCHES TO STOP MYSQL SERVER
                            # BARFING
                            batch = 100000
                            offset = 0
                            stop = 0

                            while stop == 0:
                                thisList = manyValueList[offset:offset + batch]
                                offset += batch
                                a = len(thisList)
                                cursor.executemany(sqlQuery, thisList)
                                dbConn.commit()
                                if len(thisList) < batch:
                                    stop = 1
                        i = 20
                    except:
                        pass
                if i == 10:
                    log.error('Deadlock: %s\n' % (str(e), ))
                    message = "Deadlock error"
                    raise

            else:
                message = 'MySQL write command not executed for this query: << %s >>\nThe error was: %s \n' % (sqlQuery,
                                                                                                               str(e))
                if Force == False:
                    log.error(message)
                    raise
                else:
                    log.warning(message)

        except pymysql.Warning as e:
            log.info(str(e))
        except Exception as e:
            if "truncated" in str(e):
                log.error('%s\n Here is the sqlquery:\n%s\n' %
                          (str(e), sqlQuery))
                if manyValueList:
                    log.error('... and the values:\n%s\n' % (thisList, ))
            elif "Duplicate entry" in str(e):
                log.warning('Duplicate Key error: %s\n' % (str(e), ))
                message = "duplicate key error"
            else:
                log.error(
                    'MySQL write command not executed for this query: << %s >>\nThe error was: %s \n' %
                    (sqlQuery, str(e)))
                if Force == False:
                    sys.exit(0)
                cursor.close()
                return -1
    dbConn.commit()
    # CLOSE THE CURSOR
    cOpen = True
    count = 0
    while cOpen:
        try:
            cursor.close()
            cOpen = False
        except Exception as e:
            time.sleep(1)
            count += 1
            if count == 10:
                log.warning('could not close the db cursor ' + str(e) + '\n')
                raise e
                count = 0

    log.debug('completed the ``writequery`` function')
    return message
